--[[
This file is part of Exploding numbers - a puzzle game

Exploding numbers is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
]]
EXPORT_FILE = 'exported_level'


edit = {


started = false,
saved_grid = {},
saved_current = nil,
value = 1,
history = {position = 0},


save_grid = function(self)
	local s = self.saved_grid
	in_grid(function(x, y)
		s[x] = s[x] or {}
		s[x][y] = grid[x][y]
	end)
	s.start_x, s.start_y, s.width, s.height =
		grid.start_x, grid.start_y, grid.width, grid.height
end,


restore_grid = function(self)
	grid = self.saved_grid
	self.saved_grid = {}
	self:save_grid()
end,


add_level = function(self)
	current.level = #levels + 1
	self:stop('save')
	menu.buttons = {}
	menu.order = menu.default_order
	resize_menu()
	new_message('Saved as level ' .. level_name(#levels))
end,


start = function(self)
	init_level(current.level)
	resize_grid()
	self.saved_current = current
	local edited_level = current.level
	if current.level <= levels.last_original_level then
		edited_level = #levels + 1
		new_message('Editing as level ' .. level_name(edited_level))
		levels[edited_level] = self:grid_to_level()
		levels[edited_level].solution = {unpack(levels[current.level].solution)}
		current.level = edited_level
	end
	current = {
		level = edited_level,
		goal = current.goal,
		history = {position = 0},
		selected = {}
	}
	self.history = {position = 0}
	self.started = true
	self:save_grid()
end,


stop = function(self, action)
	self:crop_grid()
	local edited_level = current.level
	local edited_goal = current.goal
	current = self.saved_current
	if action == 'delete' then
		for l = edited_level, #levels do
			-- overwrite an entry in history with later entries (if any)
			-- shifting them back
			personal_history[l] = personal_history[l + 1]
			-- same goes for levels
			levels[l] = levels[l + 1]
			-- the last one will hopefully be replaced with nil
		end
		-- can't just use table.remove(personal_history, edited_level),
		-- though it works in luajit, because the history table is not
		-- entirely populated, and the result is messy
		current.history = nil
		current.level = edited_level - 1
		new_message('Deleted level '.. level_name(edited_level))
	elseif action == 'save' then
		current.level = edited_level
		current.goal = edited_goal
		levels[edited_level] = self:grid_to_level()
	end
	init_level(current.level)
	self.started = false
	resize_grid()
	menu.order = menu.default_order
	resize_menu()
end,


click = function (self, x, y, value, new_clear)
	if not x or not y then return end
	grid[x] = grid[x] or {}
	if value == 'clear' or new_clear then
		grid = {start_x = 0, start_y = 0, height = 1, width = 1}
	elseif grid[x][y] and grid[x][y] == (value or self.value) then
		grid[x][y] = nil
	else
		grid[x][y] = value or self.value
	end
	-- fix grid size (works somehow)
	local old_width, old_height = grid.width, grid.height
	local new_end_x, new_end_y = grid.start_x + grid.width, grid.start_y + grid.height
	grid.start_x = math.min(grid.start_x, x - 1)
	grid.start_y = math.min(grid.start_y, y - 1)
	grid.width = math.max(new_end_x, x + 2) - grid.start_x
	grid.height = math.max(new_end_y, y + 2) - grid.start_y
	-- repair grid
	for x = grid.start_x, grid.width do grid[x] = grid[x] or {} end
	if not value then -- not replay
		if grid.width ~= old_width or grid.height ~= old_height then
			start_grid_animation(2) -- animate to new grid size
		else
			screen.changed = true
		end
		-- add to history
		self.history.position = self.history.position + 1
		if
			not self.history[self.history.position] or
			self.history[self.history.position][1] ~= x or
			self.history[self.history.position][2] ~= y or
			self.history[self.history.position][3] ~= (new_clear or self.value)
		then
			self.history[self.history.position] = {x, y, (new_clear or self.value)}
			-- delete rest of the history
			for s = self.history.position + 1, #self.history do
				self.history[s] = nil
			end
		end
	else
		resize_grid()
	end
	try_disabling_menu_items()
end,


undo = function(self, amount)
	local new_position = self.history.position + amount
	if new_position < 0 or new_position > #self.history then return end
	self.history.position = new_position
	self:restore_grid()
	if new_position > 0 then
		for s = 1, new_position do
			self:click(self.history[s][1], self.history[s][2], self.history[s][3])
		end
		current.selected.x = self.history[new_position][1]
		current.selected.y = self.history[new_position][2]
	end
	resize_grid()
end,


grid_to_level = function()
	if grid.height < 3 or grid.width < 3 then
		-- prevent empty grid
		return {{0}, goal = current.goal}
	end
	local l = {goal = current.goal}
	for y = grid.start_y + 1, grid.start_y + grid.height - 2 do
		l[#l + 1] = {}
		for x = grid.start_x + 1, grid.start_x + grid.width - 2 do
			l[#l][#l[#l] + 1] = (grid[x] and grid[x][y] or 0)
		end
	end
	return l
end,


export = function (self)
	self:crop_grid()
	resize_grid()
	print('Exporting level ' .. level_name(current.level))
	local l = '{\n'
	for y = grid.start_y + 1, grid.start_y + grid.height - 2 do
		l = l .. '  {'
		for x = grid.start_x + 1, grid.start_x + grid.width - 2 do
			l = l .. (x == grid.start_x + 1 and '' or ', ')
			l = l .. (grid[x] and grid[x][y] or 0)
		end
		l = l .. '},\n'
	end
	l = l .. '  goal = ' .. current.goal
	if edit.history.position == 0 then
		-- dump solution
		local sol = levels[current.level].solution
		if sol then
			l = l .. ',\n  solution = {\n    '
			for s = 1, #sol do
				l = l .. '{' .. sol[s][1] .. ', ' .. sol[s][2]
				l = l .. (s == #sol and '}' or '},')
				l = l .. (s ~= #sol and s % 7 == 0 and '\n    ' or ' ')
			end
			l = l .. '\n  }'
		end
	end
	l = l .. '\n}\n'
	io.write(l) -- write to console
	love.system.setClipboardText(l) -- to clipboard
	new_message('Exported to clipboard')
	local ok, err = love.filesystem.write(EXPORT_FILE, l) -- to file
	if not ok then new_message('File error: ' .. err, true) end
end,


crop_grid = function()
	local function delete_lines(n, rows, forwards)
		local start = 'start_' .. (rows and 'y' or 'x')
		local dimension = (rows and 'height' or 'width')
		for times = 1, n do
			if forwards then grid[start] = grid[start] + 1 end
			grid[dimension] = grid[dimension] - 1
		end
	end
	local function scan(rows, forwards)
		-- doesn't scan outer border (presumed empty)
		local v, empty_lines = {}, 0
		v[rows and 1 or 3] = grid.start_y + 1
		v[rows and 3 or 1] = grid.start_x + 1
		v[rows and 2 or 4] = grid.start_y + grid.height - 2
		v[rows and 4 or 2] = grid.start_x + grid.width - 2
		local anything = false
		for
			m =
			forwards and v[1] or v[2],
			forwards and v[2] or v[1],
			forwards and 1 or -1
		do
			for n = v[3], v[4] do
				anything = anything or (
					grid[rows and n or m] and
					grid[rows and n or m][rows and m or n]
				)
			end
			if anything then return empty_lines end
			empty_lines = empty_lines + 1
		end
		return empty_lines
	end
	for a = 1, 2 do for b = 1, 2 do
		delete_lines(scan(a == 1, b == 1), a == 1, b == 1)
	end end
end,


import = function (self)
	local function check_level(level)
		if not (level.goal and #level > 0) then
			return false
		end
		for x = 1, #level do
			if not type(level[x] == 'table') then
				return false
			end
			for y = 1, #level[x] do
				if not type(level[x][y]) == 'number' then
					return false
				end
			end
		end
		return true
	end
	local input = love.system.getClipboardText():match('%b{}')
	-- check string
	if
		not input or
		input:gsub('goal =', ''):gsub('solution =', ''):gsub('[{}, \n%-]', ''):find('%D')
	then
		new_message('Invalid input', true)
		return
	end
	local level = loadstring('return ' .. input)()
	-- check table
	if not check_level(level) then
		new_message('Invalid input', true)
		return
	end
	-- set as new level
	current.level = #levels + 1
	levels[current.level] = level
	self:stop()
	menu.buttons = {}
	menu.order = menu.default_order
	resize_menu()
	new_message('Imported as level ' .. level_name(#levels))
end,


} -- end of edit


menu.edit_start = {
	text = 'Edit',
	value = 'e',
	arrow = 'ur',
	key = 'e',
	onclick = function(self)
		settings.menu_disabled = false
		edit:start()
		menu.order = menu.edit_order
		resize_menu()
		try_disabling_menu_items()
	end
}
menu.edit_stop = {
	text = 'savE',
	value = function() return level_name(current.level) end,
	key = 'e',
	validate = function(self)
		-- change label, depending on context
		if
			edit.history[edit.history.position] and
			edit.history[edit.history.position][3] == 'clear'
		then
			self.text = 'deletE'
		elseif
			edit.history.position == 0 and
			current.goal == levels[current.level].goal
		then
			self.text = 'Exit'
		else
			self.text = 'savE'
		end
		-- center text (assume this is the last button)
		local b = screen.buttons[#screen.buttons]
		b.text_x = b.x + (settings.menu_size - menu_font_small:getWidth(self.text)) / 2
	end,
	onclick = function(self)
		if
			edit.history[edit.history.position] and
			edit.history[edit.history.position][3] == 'clear'
		then
			edit:stop('delete')
		elseif
			edit.history.position == 0 and
			current.goal == levels[current.level].goal
		then
			edit:stop() -- do nothing
		else
			edit:stop('save')
		end
	end
}
menu.edit_decrease_value = {
	text_common = 'Value',
	text_left = '\nF2 -',
	value = '',
	arrow = 'dl',
	key = 'f2',
	validate = function(self) self.disabled = edit.value <= 1 end,
	onclick = function(self)
		edit.value = math.max(1, edit.value - menu.edit_step.value)
		screen.changed = true
	end
}
menu.edit_increase_value = {
	text_right = '\n+ F3',
	value = function() return edit.value end,
	arrow = 'dr',
	common_value = true,
	key = 'f3',
	onclick = function(self)
		edit.value = edit.value + menu.edit_step.value
		screen.changed = true
	end
}
menu.edit_step = {
	text = 'Step',
	value = 1,
	key = 's',
	onclick = function(self)
		self.value = (self.value == 1) and 10 or 1
		screen.changed = true
	end
}
menu.edit_decrease_goal = {
	text_common = 'Goal',
	text_left = '\nF6 -',
	value = '',
	arrow = 'dl',
	key = 'f6',
	validate = function(self) self.disabled = current.goal <= 1 end,
	onclick = function(self)
		current.goal = math.max(1, current.goal - menu.edit_step.value)
		screen.changed = true
	end
}
menu.edit_increase_goal = {
	text_right = '\n+ F7',
	value = function() return current.goal end,
	arrow = 'dr',
	common_value = true,
	key = 'f7',
	onclick = function(self)
		current.goal = current.goal + menu.edit_step.value
		screen.changed = true
	end
}
menu.edit_clear = {
	text = 'Clear',
	value = 'F8',
	key = 'f8',
	validate = function(self) self.disabled = grid.width < 3 or grid.height < 3 end,
	onclick = function(self)
		edit:click(0, 0, nil, 'clear')
	end
}
menu.edit_rewind = {
	text = 'Rewind',
	value = 'F5',
	key = 'f5',
	validate = function(self)
		self.disabled = (edit.history.position == 0) and
			(current.goal == levels[current.level].goal)
	end,
	onclick = function(self)
		current.goal = levels[current.level].goal
		edit:undo(- edit.history.position)
	end
}
menu.edit_undo = {
	text = 'Undo',
	value = function() return edit.history.position end,
	arrow = 'dl',
	key = 'u',
	validate = function(self) self.disabled = edit.history.position <= 0 end,
	onclick = function(self) edit:undo(-1) end
}
menu.edit_redo = {
	text = 'Redo',
	value = function() return #edit.history end,
	arrow = 'dr',
	key = 'r',
	validate = function(self) self.disabled = edit.history.position >= #edit.history end,
	onclick = function(self) edit:undo(1) end
}
menu.edit_export = {
	text = 'eXport',
	value = 'x',
	key = 'x',
	validate = function(self) self.disabled = grid.width < 3 or grid.height < 3 end,
	onclick = function(self) edit:export() end
}
menu.edit_import = {
	text = 'Import',
	value = function() return level_name(#levels + 1) end,
	key = 'i',
	onclick = function() edit:import() end
}
menu.edit_add_level = {
	text = 'copY',
	value = function() return level_name(#levels + 1) end,
	key = 'y',
	validate = function(self) self.disabled = grid.width < 3 or grid.height < 3 end,
	onclick = function(self) edit:add_level() end
}
menu.edit_show = {
	text = 'Editing',
	value = function() return settings.editing and 'on' or 'off' end,
	key = 'e',
	validate = function(self) self.disabled = get_progress() < .4 end,
	onclick = function()
		settings.editing = not settings.editing
		if settings.editing then
			if menu.default_order[#menu.default_order] ~= menu.edit_start then
				menu.default_order[#menu.default_order + 1] = menu.edit_start
			end
		elseif menu.default_order[#menu.default_order] == menu.edit_start then
			menu.default_order[#menu.default_order] = nil
		end
	end
}

menu.edit_order = {
	menu.edit_decrease_value, menu.edit_increase_value,
	menu.edit_step,
	menu.edit_decrease_goal, menu.edit_increase_goal,
	menu.edit_undo, menu.edit_redo,
	menu.edit_clear, menu.edit_rewind,
	menu.edit_export, menu.edit_import,
	menu.edit_add_level,
	menu.help,
	menu.edit_stop,

}


-- one time, on load
table.insert(menu.settings_order, 1, menu.edit_show)

