--[[
Exploding numbers - a puzzle game
Copyright (C) 2022, 2023, 2024, 2025 Pajo (xpio-at-tut-dot-by)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
]]

WEB = love.system.getOS() == 'Web'

require('levels')
require('menu')
require('help')
require('saves')
require('colors')
require('edit')

-- custom run to prevent redraws when idle
function love.run()
	love.load(love.arg.parseGameArguments(arg), arg)
	-- We don't want the first frame's dt to include time taken by love.load.
	love.timer.step()
	local dt = 0
	-- Main loop time.
	return function()
		-- Process events.
		if love.event then
			love.event.pump()
			for name, a,b,c,d,e,f in love.event.poll() do
				if name == 'quit' then
					if not love.quit or not love.quit() then
						return a or 0
					end
				end
				love.handlers[name](a,b,c,d,e,f)
			end
		end
		-- Update dt, as we'll be passing it to update
		dt = love.timer.step()
		-- Call update and draw
		love.update(dt) -- will pass 0 if love.timer is disabled
		if screen.changed and love.graphics.isActive() then
			--love.graphics.origin()
			--love.graphics.clear(love.graphics.getBackgroundColor())
			love.draw()
			love.graphics.present()
			screen.changed = false
		end
		--love.timer.sleep(0.005)
		love.timer.sleep(0.02)
	end
end


function love.load(arg)
	local DEFAULT_SETTINGS = {
		save_version = SAVE_VERSION,
		color_scheme = 1,
		menu_disabled = false,
		menu_size = 60,
		min_grid = 0,
		rectangles = 0,
		solved = {n = 0},
		fullscreen = false,
	}
	animation = {grid = {}, old_values = {}}
	last_rotation_time = love.timer.getTime() -- redraw soon after start
	for a, b in pairs(arg) do
		if b == '--test' then
			get_progress = function() return 1 end
		elseif b == '--web' then
			WEB = true
		end
	end
	if WEB then table.insert(menu.default_order, menu.save_game) end
	local saved_game = load_game()
	if not saved_game then
		settings = DEFAULT_SETTINGS
		current = {}
		personal_history = {}
		init_level(1)
		love.resize()
		if #screen.buttons * settings.menu_size > screen.w then
			new_message('Scroll menu to see all buttons')
			screen.menu_drag_speed = 6
		end
	else
		local custom_levels
		settings, current, custom_levels, personal_history = unpack(saved_game)
		saved_game = nil -- release
		personal_history = personal_history or {}
		if custom_levels then
			for l = 1, #custom_levels do
				levels[#levels + 1] = custom_levels[l]
			end
		end
		if current.hash ~= level_to_hash(levels[current.level]) then
			init_level(current.level)
		end
		-- count solved levels
		settings.solved = settings.solved or {}
		settings.solved.n = 0
		for l = 1, levels.last_original_level do
			if settings.solved[level_to_hash(levels[l])] then
				settings.solved.n = settings.solved.n + 1
			end
		end
		current.history = personal_history[current.level] or {position = 0}
		if not colors[settings.color_scheme or 'not_set'] then
			-- unset for compatibilty, if the color scheme was removed
			-- default color scheme will be set in the next step
			settings.color_scheme = nil
		end
		for k, v in pairs(DEFAULT_SETTINGS) do
			settings[k] = settings[k] or v
		end
		if settings.editing then
			menu.default_order[#menu.default_order + 1] = menu.edit_start
		end
		undo(0)
		try_disabling_menu_items()
	end
	draw_playfield_shape = switch_playfield_shape(settings.rectangles)
	set_bg_color()
	toggle_fullscreen(settings.fullscreen)
end


function print_number(x, y, value, selected)
	-- determine whether to add additional marks or not
	-- also scale down long numbers
	-- other than that, scale variable is used in zooming-out animation
	-- screen.size is used for the rest
	local scale = animation.grid.font_scale or 1
	local db2, db4 = false, false
	if value == '@' then -- never underline @ sign
		selected = false
	else -- if there is an actual value
		if value > 99 then scale = scale * .66 end
		if settings.marks then
			if value == current.goal then
				-- mark goal with !
				value = value .. '!'
			else
				db4 = value % 4 == 0
				db2 = value % 2 == 0
			end
		else
			selected = false -- don't underline selection
		end
	end
	-- dots
	local function dot(offset)
		love.graphics.circle('fill',
			x + offset, y - screen.size * .33, screen.size * .035
		)
	end
	if db4 then
		dot(- screen.size * .057)
		dot(screen.size * .057)
	elseif db2 then
		dot(0)
	end
	-- the number, centered at x, y
	local xx = math.floor(x - numbers_font:getWidth(value) * scale / 2)
	local yy = math.floor(y - numbers_font:getHeight(value) * scale / 2)
	love.graphics.print(value, xx, yy, 0, scale)
	-- underline selected
	if selected then
		love.graphics.setLineWidth(screen.size * .045)
		local d = screen.size * .15
		local yy = y + 2.2 * d
		love.graphics.line(x - d, yy, x + d, yy)
	end
end


function switch_playfield_shape(rectangles)
	if rectangles == 2 then
		return function(x, y, value, selected, corner)
			set_color('naked_' .. appropriate_color(value, selected))
			if value then
				print_number(x, y, value, selected)
			else
				if selected then
					print_number(x, y, '@', true)
				elseif not corner then -- cross
					love.graphics.setLineWidth(1)
					local s = screen.size * .04
					love.graphics.line(x - s, y, x + s, y)
					love.graphics.line(x, y - s, x, y + s)
				end
			end
		end
	elseif rectangles == 1 then
		return function (x, y, value, selected, corner)
			local c = appropriate_color(value, selected)
			if not corner or selected then
				set_color(c)
				love.graphics.rectangle('fill',
					x - screen.size * .48,
					y - screen.size * .48,
					screen.size * .96,
					screen.size * .96
				)
			end
			if value then
				set_color(c .. '_numbers')
				print_number(x, y, value, selected)
			end
		end
	else -- 0
		return function (x, y, value, selected, corner)
			local c = appropriate_color(value, selected)
			if not corner or selected then
				set_color(c)
				love.graphics.circle('fill', x, y, screen.size * .48)
			end
			if value then
				set_color(c .. '_numbers')
				print_number(x, y, value, selected)
			end
		end
	end
end


function level_name(l)
	o = levels.last_original_level
	return (l <= o) and l or (l - o .. 'º')
end


function love.focus(f)
	if f then screen.changed = true end
end


function copy_history(source, position, length)
	-- can be used to copy solution to history (position should be set)
	-- or history to solution (position can be ommited)
	local destination = {position = position}
	if not source then return destination end
	for i = 1, length or #source do
		destination[#destination + 1] = {source[i][1], source[i][2]}
	end
	return destination
end


function init_level(l, replay)
	-- wrap levels
	local max_levels
	if get_progress() >= .2 then
		max_levels = #levels
	else
		max_levels = math.floor(levels.last_original_level / 2 + .5)
	end
	l = (l - 1) % max_levels + 1
	-- init new grid
	grid = {}
	grid.height = #levels[l] + 2
	grid.width = 0
	for row = 1, #levels[l] do
		grid.width = math.max(grid.width, #levels[l][row] + 2)
	end
	grid.start_x = math.ceil(grid.width / -2) - 1
	grid.start_y = math.ceil(grid.height / -2) - 1
	for x = grid.start_x, grid.width do grid[x] = {} end
	-- place selected level in grid center
	for row = 1, #levels[l] do
		for column = 1, #levels[l][row] do
			if levels[l][row][column] ~= 0 then
				local gx = grid.start_x + column
				local gy = grid.start_y + row
				grid[gx][gy] = levels[l][row][column]
			end
		end
	end
	if not replay then
		-- set some basics
		current = {
			level = l,
			goal = levels[l].goal,
			solved = settings.solved and
				settings.solved[level_to_hash(levels[l])],
		}
		-- reset screen.offsets to center the screen
		if screen then screen.offset_x, screen.offset_y = nil end
		-- init and attach history
		personal_history[l] = personal_history[l] or {position = 0}
		current.history = personal_history[l]
		current.history.position = 0 -- rewind
	end
	current.selected = {}
	if screen then
		resize_menu_disabled_button() -- goal change
		try_disabling_menu_items()
	end
end


function in_grid(f)
	for x = grid.start_x, grid.start_x + grid.width - 1 do
		for y = grid.start_y, grid.start_y + grid.height - 1 do
			local r1, r2 = f(x, y)
			if r1 or r2 then return r1, r2 end
		end
	end
end


function get_new_grid_size()
	-- disregard screen.x and screen.y (safe area offset)
	-- as translate() will have taken care of that
	local function get_offset_for_center(start, length, size, offset)
		return math.floor(offset / 2 - (start - .5 + length / 2) * size)
	end
	local size = math.floor(math.min(
		screen.w / grid.width,
		(screen.h - screen.top_margin) / grid.height
	))
	if size >= (settings.min_grid or 0) then
		return size,
			get_offset_for_center(
				grid.start_x, grid.width, size,
				screen.w
			),
			get_offset_for_center(
				grid.start_y, grid.height, size,
				screen.top_margin + screen.h
			)
	elseif not current.selected.x then
		-- at the beginning of a level or on irrelevant click
		-- set new offset or keep the things as they are
		local ox, oy = clamp_grid_offset(screen.offset_x, screen.offset_y)
		return settings.min_grid, ox, oy
	else
		-- if the minimum grid size is reached, just scroll the grid
		-- toward the selected tile
		local big_center_x = math.floor(current.selected.x * settings.min_grid)
		local big_center_y = math.floor(current.selected.y * settings.min_grid)
		local screen_center_x = math.floor(screen.w / 2)
		local screen_center_y = screen.top_margin + math.floor((screen.h - screen.top_margin) / 2)
		local new_ox, new_oy = clamp_grid_offset(
			screen_center_x - big_center_x, screen_center_y - big_center_y
		)
		return settings.min_grid, new_ox, new_oy
	end
end


function resize_font()
	if numbers_font then numbers_font:release() end
	numbers_font = love.graphics.newFont('font/medium.ttf',
		screen.size / 1.75
	)
	animation.grid.font_scale = 1
	screen.changed = true
end


function resize_grid(size, offset_x, offset_y)
	if not size then
		animation.grid.progress = nil -- stop animation
		size, offset_x, offset_y = get_new_grid_size()
	end
	screen.size = size
	screen.offset_x = offset_x
	screen.offset_y = offset_y
	if not (numbers_font and animation.grid.progress) then resize_font() end
	screen.changed = true
end


function love.resize()
	love.graphics.origin()
	screen = screen or {
		where = function(self, grid_x, grid_y)
			return self.offset_x + grid_x * self.size,
				self.offset_y + grid_y * self.size
		end,
		buttons = {},
		menu_drag = {},
		grid_drag = {}
	}
	screen.x, screen.y, screen.w, screen.h = love.window.getSafeArea()
	screen.fullscreen = love.window.getFullscreen()
	if -- check to prevent crash on a very small window
		screen.w > 2 * settings.menu_size and
		screen.h > 2 * settings.menu_size
	then
		love.graphics.translate(screen.x, screen.y)
		resize_menu()
		resize_grid()
	end
end


function start_grid_animation(duration)
	local ag = animation.grid
	-- resize font when starting animation during another animation
	if ag.progress then resize_font() end -- also resets ag.font_scale
	ag.old_size = screen.size
	ag.old_offset_x = screen.offset_x
	ag.old_offset_y = screen.offset_y
	ag.progress = 0
	ag.duration = duration
	ag.new_size, ag.new_offset_x, ag.new_offset_y = get_new_grid_size()
end


function animation_get_xy(self)
	return	self.sx * (1 - self.progress) + self.dx * self.progress,
			self.sy * (1 - self.progress) + self.dy * self.progress
end


function start_animation(sx, sy, dx, dy, value)
	animation[#animation + 1] = {
		get_xy = animation_get_xy, progress = 0, duration = .3,
		sx = sx, sy = sy, dx = dx, dy = dy, value = value,
	}
	animation.old_values[dx] = animation.old_values[dx] or {}
	animation.old_values[dx][dy] = (grid[dx] and grid[dx][dy]) or 0
end


function stop_animation(a)
	if not a then -- grid (reach final scale and offset)
		local ag = animation.grid
		ag.progress = nil -- disables animation
		resize_grid(ag.new_size, ag.new_offset_x, ag.new_offset_y)
	else -- numbers
		animation.old_values[animation[a].dx][animation[a].dy] = nil
		animation[a] = animation[#animation]
		animation[#animation] = nil
	end
end


function check_rotation(dt)
	if
		dt > 1 or -- was put to sleep
		(last_rotation_time and (love.timer.getTime() - last_rotation_time > .5))
	then
		screen.changed = true
		last_rotation_time = nil
	end
end


function love.update(dt)
	local function countdown(t, v, f)
		if t and t[v] then
			t[v] = t[v] - 1
			if t[v] <= 0 then f() end
		end
	end
	local ag, progress = animation.grid, animation.grid.progress
	if progress then
		screen.changed = true
		resize_grid(
			ag.old_size * (1 - progress) + ag.new_size * progress,
			ag.old_offset_x * (1 - progress) + ag.new_offset_x * progress,
			ag.old_offset_y * (1 - progress) + ag.new_offset_y * progress
		)
		ag.font_scale = (1 - progress) + ag.new_size / ag.old_size * progress
		ag.progress = progress + dt / ag.duration
		if ag.progress >= 1 then stop_animation() end
	end
	for a = 1, #animation do
		if not animation[a] then break end
		screen.changed = true
		animation[a].progress = animation[a].progress + dt / animation[a].duration
		if animation[a].progress >= 1 then stop_animation(a) end
	end
	if screen.menu_drag_speed then
		shift_menu(screen.buttons.dx + screen.menu_drag_speed)
		screen.menu_drag_speed = screen.menu_drag_speed / 1.3
		if math.abs(screen.menu_drag_speed) < .1 then screen.menu_drag_speed = nil end
	end
	if screen.messages then
		countdown(screen.messages[#screen.messages], 'delay', function()
			screen.messages[#screen.messages] = nil
			if #screen.messages == 0 then screen.messages = nil end
			screen.changed = true
		end)
	end
	countdown(screen, 'zoom_indicator_delay', function()
		screen.zoom_indicator_delay = nil
		screen.changed = true
	end)
	countdown(screen.mousemoved, 'delay', function()
		screen.mousemoved = {0}
	end)
	check_rotation(dt)
end


function love.displayrotated()
	last_rotation_time = love.timer.getTime()
end


function love.visible(v)
	last_rotation_time = v and love.timer.getTime()
end


function explode(x, y, replay)
	local function victory()
		local old_p = get_progress()
		local h = level_to_hash(levels[current.level])
		if
			not settings.solved[h]
			and current.history == personal_history[current.level]
		then
			settings.solved[h] = true
			if current.level <= levels.last_original_level then
				settings.solved.n = settings.solved.n + 1
			end
			current.solved = true
		end
		local new_p = get_progress()
		for a = .2, 1, .2 do if old_p < a and new_p >= a then
			new_message(draw_summary, false, 400)
		end end
		-- dump_solution
		if current.level and current.history then
			-- drop any history entries beyond the solution
			for s = current.history.position + 1, #current.history do
				current.history[s] = nil
			end
			-- prepare to print
			local steps = #current.history ..
				(#current.history == 1 and ' step' or ' steps')
			-- check if it is a new or better solution
			if
				not levels[current.level].solution or
				#levels[current.level].solution > #current.history
			then
				levels[current.level].solution = copy_history(current.history)
				new_message('New solution\n' .. steps)
				current.already_won = false
			end
			-- in any case print the solution in console
			if not current.already_won then
				print('Level: ' .. level_name(current.level))
				for s = 1, #current.history do
					io.write('{' .. current.history[s][1] .. ', ' .. current.history[s][2] .. '}, ')
				end
				print('\n' .. steps)
			end
		end
		-- end of dump_solution part
		current.already_won = true
	end -- of victory ()
	local function in_directions(x, y, f)
		for a = 1, 7, 2 do
			local b, c = select(a, -1, 0, 1, 0, 0, -1, 0, 1)
			f(x + b, y + c)
		end
	end
	local function place(x, y, value)
		grid[x] = grid[x] or {}
		grid[x][y] = (grid[x][y] or 0) + value
		local grid_expanded = false
		if x - 1 < grid.start_x then
			grid.start_x = x - 1
			grid.width = grid.width + 1
			grid_expanded = true
		end
		if x + 1 > grid.width + grid.start_x - 1 then
			grid.width = grid.width + 1
			grid_expanded = true
		end
		if y - 1 < grid.start_y then
			grid.start_y = y - 1
			grid.height = grid.height + 1
			grid_expanded = true
		end
		if y + 1 > grid.height + grid.start_y - 1 then
			grid.height = grid.height + 1
			grid_expanded = true
		end
		return grid_expanded
	end
	if not (x and y and grid[x]) then return end
	local grid_expanded = false
	local level_changed = false
	if grid[x][y] and grid[x][y] == current.goal then
		init_level(current.level + 1)
		resize_grid()
		return
	elseif grid[x][y] and grid[x][y] % 4 == 0 then
		in_directions(x, y, function (xx, yy)
			if not replay then start_animation(x, y, xx, yy, grid[x][y] / 4) end
			grid_expanded = place(xx, yy, grid[x][y] / 4) or grid_expanded
		end)
		grid[x][y] = nil
		level_changed = true
	elseif not grid[x][y] then
		in_directions(x, y, function (xx, yy)
			if grid[xx] and grid[xx][yy] and grid[xx][yy] % 2 == 0 then
				local new_x, new_y = 2 * xx - x, 2 * yy - y
				if not replay then start_animation(xx, yy, new_x, new_y, grid[xx][yy] / 2) end
				grid[xx][yy] = grid[xx][yy] / 2
				grid_expanded = place(new_x, new_y, grid[xx][yy]) or grid_expanded
				level_changed = true
			end
		end)
	else
		current.selected.x, current.selected.y = nil, nil
	end
	if grid_expanded then
		-- repair grid
		for x = grid.start_x, grid.width do grid[x] = grid[x] or {} end
		if not replay and not screen.grid_drag[1] then
			start_grid_animation(2) -- animate to new grid size
		end
	end
	if replay then
		-- show clicks when playing back history
		current.selected.x, current.selected.y = x, y
	elseif level_changed then
		current.history.position = current.history.position + 1
		-- add click and clear the rest of the history
		-- unless the same click is repeated
		if
			not current.history[current.history.position] or
			current.history[current.history.position][1] ~= x or
			current.history[current.history.position][2] ~= y
		then
			current.history[current.history.position] = {x, y}
			for a = current.history.position + 1, #current.history do
				current.history[a] = nil
			end
		end
		-- these checks are not done at level start, only afterwards
		try_disabling_menu_items()
	elseif
		current.solved and
		x == grid.start_x + grid.width - 1 and
		y == grid.start_y + grid.height - 1
	then
		new_message(draw_summary, false, 400)
	end
	-- check victory
	if not replay and not current.already_won then
		in_grid(function(x, y)
			if grid[x][y] == current.goal then
				victory() -- sets current.already_won
				return true -- exits in_grid()
			end
		end)
	end
end


function draw_message(m)
	local PAD = 15
	local w, lines = menu_font_large:getWrap(m.message, screen.w)
	local h = menu_font_large:getHeight(lines[1]) * #lines
	local x = math.floor((screen.w - w) / 2)
	local y = math.floor((screen.h - h) / 2)
	local name = m.warning and 'warning' or 'message'
	love.graphics.setColor(0, 0, 0, .5)
	love.graphics.rectangle('fill', x, y, w + 2 * PAD, h + 2 * PAD)
	set_color(name .. '_bg')
	love.graphics.rectangle('fill', x - PAD, y - PAD, w + 2 * PAD, h + 2 * PAD)
	set_color(name .. '_text')
	love.graphics.setFont(menu_font_large)
	love.graphics.printf(m.message, 0, y, screen.w, 'center')
end


function get_progress()
	return settings.solved.n / levels.last_original_level
end


function draw_summary()
	local function print_lock(x, y, locked, hs)
		local lx = locked and .62 or .99
		local ly = locked and .74 or .70
		love.graphics.print('U', x + hs * lx, y + hs * ly, math.pi)
		love.graphics.rectangle('fill', x, y + hs * .32, hs * .62, hs * .52)
	end
	local PAD = 15
	local txt1 = settings.solved.n .. ' of ' ..
		levels.last_original_level .. ' levels solved'
	local txt2 = 'all levels\neditor\nshapes\nall color schemes'
	local w1, lines1 = menu_font_large:getWrap(txt1, screen.w)
	local h_single = menu_font_large:getHeight(lines1[1])
	local ox = 2 * h_single
	local w2, lines2 = menu_font_large:getWrap(txt2, screen.w - ox)
	w2 = w2 + ox
	local w = math.max(w1, w2)
	local h = h_single * (#lines1 + #lines2 + 1)
	local x = math.floor((screen.w - w) / 2)
	local y = math.floor((screen.h - h) / 2)
	love.graphics.setColor(0, 0, 0, .5)
	love.graphics.rectangle('fill', x, y, w + 2 * PAD, h + 2 * PAD)
	set_color('message_bg')
	love.graphics.rectangle('fill', x - PAD, y - PAD, w + 2 * PAD, h + 2 * PAD)
	set_color('message_text')
	love.graphics.setFont(menu_font_large)
	love.graphics.printf(txt1, x, y, w, 'center')
	local y2 = y + h_single * (#lines1 + 1)
	love.graphics.printf(txt2, x + ox, y2, w - h_single, 'left')
	for a = 1, 4 do
		print_lock(x + h_single * .6 , y2 + h_single * (a - 1),
			get_progress() < a * .2,
			h_single
		)
	end
	-- draw bar
	local progress = w * get_progress()
	love.graphics.setLineWidth(1)
	love.graphics.rectangle('line',
		x, y + h_single * (#lines1 + .3),
		w, h_single * .3
	)
	love.graphics.rectangle('fill',
		x, y + h_single * (#lines1 + .3),
		progress, h_single * .3
	)
	for a = .2, .8, .2 do
		local bar_pos = x + w * a
		set_color(bar_pos < x + progress and 'message_bg' or 'message_text')
		love.graphics.line(
			bar_pos, y + h_single * (#lines1 + .3),
			bar_pos, y + h_single * (#lines1 + .6)
		)
	end
end


function draw_grid()
	local function print_check_mark(x, y, selected)
		local scale = animation.grid.font_scale or 1
		set_color(
			settings.rectangles == 2 and 'naked_indicator'
			or selected and 'selected_numbers' or 'indicator'
		)
		love.graphics.print(
			'L',
			math.floor(x + numbers_font:getWidth('L') * scale * 1.2),
			math.floor(y - numbers_font:getHeight('L') * scale * .2),
			math.pi / 4, -scale ,scale
		)
	end
	love.graphics.setFont(numbers_font)
	in_grid(function(x, y)
		local value = (animation.old_values[x] and animation.old_values[x][y]) or grid[x][y]
		value = value ~= 0 and value -- reset to false if 0
		local screen_x, screen_y = screen:where(x, y)
		local selected = current.selected.x == x and current.selected.y == y
		draw_playfield_shape(
			screen_x, screen_y, value, selected,
			(x == grid.start_x or x == grid.start_x + grid.width - 1) and
			(y == grid.start_y or y == grid.start_y + grid.height - 1)
		)
	end)
	for a = 1, #animation do
		local screen_x, screen_y = screen:where(animation[a]:get_xy())
		set_color(
			(settings.rectangles == 2 and 'naked_' or '') ..
			appropriate_color(animation[a].value) ..
			(settings.rectangles == 2 and '' or '_numbers')
		)
		print_number(screen_x, screen_y, animation[a].value)
	end
	if current.solved then
		local corner_x = grid.start_x + grid.width - 1
		local corner_y = grid.start_y + grid.height - 1
		local selected = current.selected.x == corner_x and
			current.selected.y == corner_y
		if settings.rectangles == 2 and selected then return end
		-- print achievement
		local screen_x, screen_y = screen:where(corner_x, corner_y)
		print_check_mark(screen_x, screen_y, selected)
	end
end


function draw_min_zoom_indicator(selected)
	if not (
		screen.zoom_indicator_delay and screen.size > settings.min_grid
	) then return end
	set_color(
		settings.rectangles == 2 and 'naked_indicator'
		or selected and 'selected_numbers' or 'indicator'
	)
	love.graphics.setLineWidth(3)
	local x, y = screen:where(grid.start_x, grid.start_y)
	local s = settings.min_grid
	if settings.rectangles == 0 then
		love.graphics.circle('line', x, y, s / 2)
	else
		love.graphics.rectangle('line', x - s / 2, y - s / 2, s, s)
	end
end


function love.draw()
	love.graphics.clear(colors.bg_color)
	draw_grid()
	draw_min_zoom_indicator(
		current.selected.x == grid.start_x and
		current.selected.y == grid.start_y
	)
	draw_menu()
	if screen.messages then
		local m = screen.messages[#screen.messages]
		if type(m.message) == 'function' then
			m.message()
		else
			draw_message(m)
		end
	end
end


function new_message(message, warning, delay)
	screen.messages = screen.messages or {}
	table.insert(screen.messages, 1, {
		message = message,
		warning = warning,
		delay = delay or 150
	})
	screen.changed = true
end


function undo(amount)
	current.already_won = false -- demand recheck
	local new_position = current.history.position + amount
	if new_position < 0 or new_position > #current.history then return end
	current.history.position = new_position
	init_level(current.level, true)
	for s = 1, new_position do
		explode(current.history[s][1], current.history[s][2], true)
	end
	if not screen then love.resize() else resize_grid() end
end


function love.keypressed(k)
	local lk_id = love.keyboard.isDown
	if make_last_message_expire() then return end
	local function move_cursor(x, y)
		current.selected.x = (
			((current.selected.x or 0) + x - grid.start_x) %
			grid.width + grid.start_x
		)
		current.selected.y = (
			((current.selected.y or 0) + y - grid.start_y) %
			grid.height + grid.start_y
		)
		if (screen.size <= (settings.min_grid or 0)) then
			start_grid_animation(1)
		end
		screen.changed = true
	end
	if k == 'h' or k == 'left' or k == 'kp4' then
		if lk_id('lshift') or lk_id('rshift') then
			local offset = screen.buttons.dx % settings.menu_size
			shift_menu(screen.buttons.dx + settings.menu_size - offset)
		else
			move_cursor(-1, 0)
		end
	elseif k == 'j' or k == 'down' or k == 'kp2' then
		move_cursor(0, 1)
	elseif k == 'k' or k == 'up' or k == 'kp8' then
		move_cursor(0, -1)
	elseif k == 'l' or k == 'right' or k == 'kp6' then
		if lk_id('lshift') or lk_id('rshift') then
			local offset = screen.buttons.dx % settings.menu_size
			offset = offset == 0 and settings.menu_size or offset
			shift_menu(screen.buttons.dx - offset)
		else
			move_cursor(1, 0)
		end
	elseif k == 'kp1' then
		move_cursor(-1, 1)
	elseif k == 'kp3' then
		move_cursor(1, 1)
	elseif k == 'kp7' then
		move_cursor(-1, -1)
	elseif k == 'kp9' then
		move_cursor(1, -1)
	elseif
		k == 'space' or k == 'return' or k == 'kpenter' or k == 'kp5'
	then
		if edit.started then
			edit:click(current.selected.x, current.selected.y)
		else
			local csx, csy = current.selected.x, current.selected.y
			explode(csx, csy)
			current.selected.x, current.selected.y = csx, csy
		end
	elseif k == 'escape' then
		exit_menu_or_quit()
	elseif k == 'q' and not WEB then
		love.event.push('quit')
	elseif k == 'c' then
		menu['colors_' .. (
			(lk_id('lshift') or lk_id('rshift')) and 'prev' or 'next'
		)]:onclick()
	else -- check menu keys
		for m = 1, #menu.order do
			if k == menu.order[m].key and not menu.order[m].disabled then
				menu.order[m]:onclick()
				try_disabling_menu_items()
				return
			end
		end
	end
end


function isin(xx, yy, x, y, w, h)
	return xx >= x and xx <= (x + w) and yy >= y and yy <= (y + h)
end


function love.mousereleased(mx, my)
	mx, my = love.graphics.inverseTransformPoint(mx, my)
	if screen.menu_drag[1] and (not screen.mousemoved[1]) then
		screen.menu_drag = {}
		screen.mousemoved = {}
		-- click menu
		local buttons = screen.buttons
		local menu_w = #screen.buttons * settings.menu_size
		for a = 1, #menu.order do
			if
				(
					isin(
						mx - buttons.dx, my,
						buttons[a].x, buttons[a].y,
						buttons[a].w, buttons[a].h
					)
					or (
						screen.w < menu_w
						and isin(
							menu_w + mx - buttons.dx, my,
							buttons[a].x, buttons[a].y,
							buttons[a].w, buttons[a].h
						)
					)
				)
				and menu.order[a].onclick
				and not menu.order[a].disabled
			then
				menu.order[a]:onclick(a)
				try_disabling_menu_items()
				return
			end
		end
	elseif screen.grid_drag[1] and (not screen.mousemoved[1]) then
		grid_click(screen.grid_drag[1], screen.grid_drag[2])
		screen.grid_drag = {}
		screen.mousemoved = {}
	end
	if screen.menu_drag[1] then
		screen.menu_drag_speed = screen.mousemoved[1]
	end
	screen.menu_drag = {}
	screen.grid_drag = {}
	screen.mousemoved = {}
end


function love.mousemoved(mx, my, dx, dy)
	if screen.touch_distance then -- when resizing with pinch
		screen.grid_drag[1] = nil -- prevents wrong offset when ending pinch
		return -- prevents quick offset-flickering
	end
	mx, my = love.graphics.inverseTransformPoint(mx, my)
	local d
	if screen.grid_drag[1] then
		d = screen.grid_drag
	elseif screen.menu_drag[1] then
		d = screen.menu_drag
	end
	if d then
		if
			math.abs(d[1] - mx) > 5 or
			math.abs(d[2] - my) > 5
		then
			screen.mousemoved[1] = dx * 1.5
			screen.mousemoved[2] = dy * 1.5
		end
		if d == screen.menu_drag then
			shift_menu(screen.buttons.dx + dx)
		else
			screen.offset_x, screen.offset_y = clamp_grid_offset(
				screen.offset_x + dx, screen.offset_y + dy
			)
			screen.changed = true
		end
	end
end


function clamp_grid_offset(ox, oy)
	-- extreme offsets, where one side of the grid touches the safe area's edge
	local l = math.floor((.5 - grid.start_x) * settings.min_grid)
	local r = screen.w - math.floor((grid.start_x + grid.width - .5) * settings.min_grid)
	local u = screen.top_margin + math.floor((.5 - grid.start_y) * settings.min_grid)
	local d = screen.h - math.floor((grid.start_y + grid.height - .5) * settings.min_grid)
	-- center at the selected tile if the grid doesn't fit the screen
	-- and if there is a selected tile (only ox checked)
	-- or else, return grid center
	return
		(ox and (r - l < 0) and math.max(r, math.min(l, ox))) or
			((l + r) / 2)
		,
		(ox and (d - u < 0) and math.max(d, math.min(u, oy)) or
			((u + d) / 2))
end


function love.wheelmoved(mx, my)
	local function biggest_of(mx, my)
		return ((math.abs(mx) > math.abs(my)) and mx or my)
	end
	if
		love.keyboard.isDown('lshift') or love.keyboard.isDown('rshift')
	then
		mx, my = my, mx
	end
	if not settings.menu_disabled and love.mouse.getY() <= settings.menu_size then
		if
			love.keyboard.isDown('lctrl') or love.keyboard.isDown('rctrl')
		then -- maximize menu buttons
			settings.menu_size = math.max(60,
				math.min(100, settings.menu_size + biggest_of(mx, my) * 5)
			)
			resize_menu_fonts()
			love.resize()
		else
			shift_menu(screen.buttons.dx + biggest_of(mx, my) * settings.menu_size)
		end
	elseif
		love.keyboard.isDown('lctrl') or love.keyboard.isDown('rctrl')
	then
		local amount = biggest_of(mx, my) * 10
		-- prevent dragging while zooming out
		-- as it creates bad offsets
		if amount < 0 then screen.grid_drag[1] = nil end
		resize_min_grid_by(amount)
		resize_grid()
		screen.zoom_indicator_delay = 100
	elseif screen.size == settings.min_grid then
		screen.offset_x, screen.offset_y = clamp_grid_offset(
			screen.offset_x + mx * 10, screen.offset_y + my * 10
		)
		screen.changed = true
	end
end


function grid_click(mx, my)
	current.selected.x, current.selected.y = in_grid(function(x, y)
		local screen_x, screen_y = screen:where(x, y)
		if isin(
			mx, my,
			screen_x - screen.size / 2, screen_y - screen.size / 2,
			screen.size, screen.size
		) then
			return x, y
		end
	end)
	if edit.started then
		edit:click(current.selected.x, current.selected.y)
	else
		explode(current.selected.x, current.selected.y)
	end
	screen.changed = true
end


function make_last_message_expire() -- doesn't work for warnings
	if
		screen.messages and screen.messages[#screen.messages]
		and not screen.messages[#screen.messages].warning
	then
		screen.messages[#screen.messages].delay = 0
		return true
	end
end


function love.mousepressed(mx, my, button)
	if make_last_message_expire() then return end
	mx, my = love.graphics.inverseTransformPoint(mx, my)
	-- check menu
	local bmd = screen.menu_disabled_button
	if not settings.menu_disabled and my <= settings.menu_size then
		-- begin dragging menu
		screen.menu_drag = {mx, my}
		screen.mousemoved = {}
		return
	elseif isin(mx, my, bmd.x, bmd.y, bmd.ww + bmd.h / 4, bmd.h) then
		menu.goal:onclick()
		return
	elseif screen.size == settings.min_grid then
		-- begin dragging grid
		screen.grid_drag = {mx, my}
		screen.mousemoved = {delay = 25}
		return
	end
	-- if no zoom, click immediately
	grid_click(mx, my)
end


function get_touch_distance(t)
	if #t ~= 2 then return end
	local x1, y1 = love.touch.getPosition(t[1])
	local x2, y2 = love.touch.getPosition(t[2])
	return math.sqrt((x1 - x2) ^ 2 + (y1 - y2) ^ 2)
end


function love.touchpressed()
	screen.touch_distance = get_touch_distance(love.touch.getTouches())
end


function love.touchreleased()
	screen.touch_distance = get_touch_distance(love.touch.getTouches())
end


function love.touchmoved()
	local new_distance = get_touch_distance(love.touch.getTouches())
	if not (screen.touch_distance and new_distance) then
		screen.touch_distance = nil
		return
	end
	local d = math.ceil(new_distance - screen.touch_distance - .5)
	screen.touch_distance = new_distance
	if d == 0 then return end
	resize_min_grid_by(d)
	resize_grid()
	screen.zoom_indicator_delay = 100
end


function resize_min_grid_by(d)
	settings.min_grid = math.min(200, math.max(0, (settings.min_grid or 0) + d))
end


function toggle_fullscreen(fs)
	if fs == nil then fs = not screen.fullscreen end
	if fs then
		-- remember window size before going fullscreen
		screen.window_w, screen.window_h, screen.window_flags =
			love.window.getMode()
		love.window.setFullscreen(fs)
		screen.fullscreen = love.window.getFullscreen()
	else
		if screen.window_flags and love.system.getOS() ~= 'Android' then
			love.window.setMode(
				screen.window_w, screen.window_h, screen.window_flags
			)
		else
			love.window.setFullscreen(fs)
		end
		love.resize() -- also sets screen.fullscreen
	end
	settings.fullscreen = screen.fullscreen
end

function love.quit()
	save_game()
	if love.system.getOS() == 'Android' then
		os.exit() -- prevents some crashes
	end
end
