--[[
This file is part of Exploding numbers - a puzzle game

Twinnings is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
]]


SAVE_FILE = 'save'
SAVE_VERSION = 3


local function serialize(o, data) -- adapted from "Programming in Lua"
	data = data or ''
	if type(o) == 'number' then
		data = data .. o
	elseif type(o) == 'string' then
		data = data .. string.format('%q', o)
	elseif type(o) == 'boolean' then
		if o then data = data .. 'true' else data = data .. 'false' end
	elseif type(o) == 'table' then
		data = data .. '{\n'
	for k,v in pairs(o) do
		if type(k) == 'number' then
			data = data .. '  [' .. k .. '] = '
		elseif type(k) == 'string' and #k == 28 then -- probably hash with = and \
			data = data .. '  ["' .. k .. '"] = '
		else
			data = data .. '  ' .. k .. ' = '
		end
		data = data .. serialize(v)
		data = data .. ',\n'
	end
		data = data .. '}\n'
	else
		error('cannot serialize a ' .. type(o))
	end
	return data
end


function level_to_hash(l)
	-- a more primitive 'serialize' is needed, because it seems that
	-- hash algorithms are still buggy with long strings
	local d = l.goal
	for x = 1, #l do
		d = d .. 'r'
		for y = 1, #l[x] do
			d = d .. l[x][y] .. ','
		end
	end
	return love.data.encode('string', 'base64', love.data.hash('sha1', d))
end


function del_game()
	love.filesystem.remove(SAVE_FILE)
	love.filesystem.remove(EXPORT_FILE)
end


function load_game()
	local function load_file()
		local saved_game, err = love.filesystem.load(SAVE_FILE)
		if not err then
			return saved_game()
		end
	end
	local saved_game = love.filesystem.getInfo(SAVE_FILE) and load_file()
	if
		saved_game and type(saved_game[1]) == 'table' and
		saved_game[1].save_version == SAVE_VERSION and
		type(saved_game[2]) == 'table' and (
			levels[saved_game[2].level] or (
				type(saved_game[3]) == 'table' and
				type(saved_game[3][saved_game[2].level - #levels]) == 'table'
			)
		)
	then
		return saved_game
	else
		return false
	end
end


function save_game()
	local function save_file(t)
		local ok, err = love.filesystem.write(SAVE_FILE, 'return ' .. serialize(t))
		if not ok then new_message('File error: ' .. err, true) end
	end
	local custom_levels = {}
	for l = levels.last_original_level + 1, #levels do
		custom_levels[#custom_levels + 1] = levels[l]
	end
	if current then
		current.hash = level_to_hash(levels[current.level])
		-- detach to avoid duplication
		-- current.history will be linked with personal_history on load
		current.history = nil
		save_file{settings, current, custom_levels, personal_history}
		-- restore after save (relevant in web platform's mid-game save)
		current.history = personal_history[current.level] or {position = 0}
	end
end
